/******************************************************************************
*                                                                             *
*    DirectoryFileList.cpp                  Copyright(c) 2005-2009 itow,y.    *
*                                                                             *
******************************************************************************/

/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdlib.h>
#include <windows.h>
#include <shlobj.h>
#include <tchar.h>
#include "DirectoryFileList.h"
#include "Util.h"




FileInfo::FileInfo(LPCTSTR pszFileName,ULONGLONG FileSize,const FILETIME &LastWriteTime)
{
	m_pszFileName=new TCHAR[lstrlen(pszFileName)+1];
	lstrcpy(m_pszFileName,pszFileName);
	m_FileSize=FileSize;
	m_LastWriteTime=LastWriteTime;
	m_pszTypeName=NULL;
	m_hIcon=NULL;
}


FileInfo::~FileInfo()
{
	delete [] m_pszFileName;
	delete [] m_pszTypeName;
	if (m_hIcon!=NULL)
		DestroyIcon(m_hIcon);
}


LPCTSTR FileInfo::GetTypeName()
{
	if (m_pszTypeName==NULL) {
		SHFILEINFO shfi;

		if (SHGetFileInfo(m_pszFileName,0,&shfi,sizeof(SHFILEINFO),
								SHGFI_TYPENAME | SHGFI_USEFILEATTRIBUTES)) {
			m_pszTypeName=new TCHAR[lstrlen(shfi.szTypeName)+1];
			lstrcpy(m_pszTypeName,shfi.szTypeName);
		}
	}
	return m_pszTypeName;
}


HICON FileInfo::GetIcon()
{
	if (m_hIcon==NULL) {
		SHFILEINFO shfi;

		if (SHGetFileInfo(m_pszFileName,0,&shfi,sizeof(SHFILEINFO),
					SHGFI_ICON | SHGFI_SMALLICON | SHGFI_USEFILEATTRIBUTES))
			m_hIcon=shfi.hIcon;
	}
	return m_hIcon;
}




DirectoryFileList::DirectoryFileList()
{
	m_ppList=NULL;
	m_ListLength=NULL;
	m_NumFiles=0;
	m_szDirectory[0]='\0';
}


DirectoryFileList::~DirectoryFileList()
{
	Clear();
}


void DirectoryFileList::Clear()
{
	if (m_ppList!=NULL) {
		for (int i=m_NumFiles-1;i>=0;i--)
			delete m_ppList[i];
		free(m_ppList);
		m_ppList=NULL;
		m_ListLength=0;
		m_NumFiles=0;
	}
	m_szDirectory[0]='\0';
}


bool DirectoryFileList::AddFile(const WIN32_FIND_DATA *pwfd)
{
	if (m_NumFiles==m_ListLength) {
		int NewLength=m_ListLength==0?32:m_ListLength*2;
		FileInfo **ppNewList;

		ppNewList=(FileInfo**)realloc(m_ppList,
							NewLength*sizeof(FileInfo*));
		if (ppNewList==NULL)
			return false;
		m_ppList=ppNewList;
		m_ListLength=NewLength;
	}
	m_ppList[m_NumFiles]=new FileInfo(pwfd->cFileName,
		((ULONGLONG)pwfd->nFileSizeHigh<<32)|pwfd->nFileSizeLow,
		pwfd->ftLastWriteTime);
	m_NumFiles++;
	return true;
}


bool DirectoryFileList::FindFiles(LPCTSTR pszDirectory,LPCTSTR pszMasks)
{
	HANDLE hFind;
	WIN32_FIND_DATA wfd;
	LPCTSTR p;
	int Prev;
	int i;
	TCHAR szMask[MAX_PATH];
	TCHAR szFileName[MAX_PATH];

	Clear();
	p=pszMasks;
	Prev=0;
	while (*p!='\0') {
		for (i=0;*p!='\0' && *p!=';';i++)
			szMask[i]=*p++;
		szMask[i]='\0';
		MergeFileName(szFileName,pszDirectory,szMask);
		hFind=FindFirstFile(szFileName,&wfd);
		if (hFind!=INVALID_HANDLE_VALUE) {
			do {
				if ((wfd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)==0) {
					for (i=0;i<Prev;i++)
						if (lstrcmpi(m_ppList[i]->GetFileName(),wfd.cFileName)==0)
							break;
					if (i==Prev) {
						if (!AddFile(&wfd)) {
							Clear();
							FindClose(hFind);
							return false;
						}
					}
				}
			} while (FindNextFile(hFind,&wfd));
			FindClose(hFind);
		}
		if (*p==';')
			p++;
		Prev=m_NumFiles;
	}
	lstrcpy(m_szDirectory,pszDirectory);
	return true;
}


void DirectoryFileList::GetDirectory(LPTSTR pszDirectory)
{
	lstrcpy(pszDirectory,m_szDirectory);
}


bool DirectoryFileList::AddFile(LPCTSTR pszFileName)
{
	TCHAR szFilePath[MAX_PATH];
	HANDLE hFind;
	WIN32_FIND_DATA wfd;

	if (GetFileIndex(pszFileName)>=0)
		return true;
	if (m_szDirectory[0]!='\0')
		MergeFileName(szFilePath,m_szDirectory,pszFileName);
	else
		lstrcpy(szFilePath,pszFileName);
	hFind=FindFirstFile(szFilePath,&wfd);
	if (hFind==INVALID_HANDLE_VALUE)
		return false;
	FindClose(hFind);
	return AddFile(&wfd);
}


bool DirectoryFileList::GetFile(int Index,LPTSTR pszFileName)
{
	if (Index<0 || Index>=m_NumFiles)
		return false;
	MergeFileName(pszFileName,m_szDirectory,m_ppList[Index]->GetFileName());
	return true;
}


int DirectoryFileList::GetFileIndex(LPCTSTR pszFileName)
{
	int i;

	for (i=0;i<m_NumFiles;i++) {
		if (lstrcmpi(m_ppList[i]->GetFileName(),pszFileName)==0)
			return i;
	}
	return -1;
}


static int CompareName(FileInfo *pFile1,FileInfo *pFile2)
{
	return lstrcmpi(pFile1->GetFileName(),pFile2->GetFileName());
}


static int CompareNameNumber(FileInfo *pFile1,FileInfo *pFile2)
{
	return StringNumberCompare(pFile1->GetFileName(),pFile2->GetFileName(),true);
}


static int CompareType(FileInfo *pFile1,FileInfo *pFile2)
{
	if (pFile1->GetTypeName()==NULL || pFile2->GetTypeName()==NULL)
		return 0;
	return lstrcmpi(pFile1->GetTypeName(),pFile2->GetTypeName());
}


static int CompareExtension(FileInfo *pFile1,FileInfo *pFile2)
{
	return lstrcmpi(GetFileExtensionPtr(pFile1->GetFileName()),
					GetFileExtensionPtr(pFile2->GetFileName()));
}


static int CompareSize(FileInfo *pFile1,FileInfo *pFile2)
{
	if (pFile1->GetFileSize()<pFile2->GetFileSize())
		return -1;
	if (pFile1->GetFileSize()>pFile2->GetFileSize())
		return 1;
	return 0;
}


static int CompareLastWriteTime(FileInfo *pFile1,FileInfo *pFile2)
{
	return CompareFileTime(&pFile1->GetLastWriteTime(),
						   &pFile2->GetLastWriteTime());
}


void DirectoryFileList::SortSub(FileInfo **ppFirst,FileInfo **ppLast,
								CompareFileInfoFunc pCompare,bool fDescending)
{
	FileInfo **p,**q;
	FileInfo *pKey=ppFirst[(ppLast-ppFirst)/2];
	int Cmp;

	p=ppFirst;
	q=ppLast;
	while (p<=q) {
		while (true) {
			Cmp=(*pCompare)(*p,pKey);
			if (Cmp==0)
				Cmp=CompareName(*p,pKey);
			if (fDescending)
				Cmp=-Cmp;
			if (Cmp>=0)
				break;
			p++;
		}
		while (true) {
			Cmp=(*pCompare)(*q,pKey);
			if (Cmp==0)
				Cmp=CompareName(*q,pKey);
			if (fDescending)
				Cmp=-Cmp;
			if (Cmp<=0)
				break;
			q--;
		}
		if (p<=q) {
			FileInfo *pTmp;

			pTmp=*p;
			*p=*q;
			*q=pTmp;
			p++;
			q--;
		}
	}
	if (q>ppFirst)
		SortSub(ppFirst,q,pCompare,fDescending);
	if (p<ppLast)
		SortSub(p,ppLast,pCompare,fDescending);
}


/*
	t@Cבւ
*/
void DirectoryFileList::Sort(SortType Type,bool fDescending/*=false*/)
{
	if (m_NumFiles>1) {
		static const CompareFileInfoFunc pCompareFuncList[] = {
			CompareName,
			CompareNameNumber,
			CompareType,
			CompareExtension,
			CompareSize,
			CompareLastWriteTime
		};

		SortSub(&m_ppList[0],&m_ppList[m_NumFiles-1],pCompareFuncList[Type],
																fDescending);
	}
}


HICON DirectoryFileList::GetFileIcon(int Index)
{
	if (Index<0 || Index>=m_NumFiles)
		return NULL;
	return m_ppList[Index]->GetIcon();
}
